/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"
#include "PartIdList.h"
#include "Iterator.h"

//////////////////////////////////////////////////////////////////
//
// Reorganization Test #2 - Randomly delete and reinsert as many
// atomic parts per composite part as each one has, interleaving
// reorganizations of all of the composite parts, to test out the
// deletion/reallocation handling and performance characteristics
// of the system
//
//////////////////////////////////////////////////////////////////

extern char*	types[NumTypes];
extern o_4b	debugMode;

static Iterator	partI;

//////////////////////////////////////////////////////////////////
//
// Routine to do Reorg2
//
//////////////////////////////////////////////////////////////////

o_4b reorg2()
{
    // for as many atomic parts as each composite part contains,
    // iterate through all the composite parts, deleting and then
    // recreating one of the atomic parts from its atomic parts graph
    // (and count the number of atomic parts processed in this way)

    o_4b		count = 0;
    CompositePart*	cp;
    PPredicate		pred = NULL_PREDICATE;

    for (o_word i = 0; i < NumAtomicPerComp; i++) 
    {
	// Iterate over all composite parts.
	partI.select(Composite, pred);
	for (cp = (Link<CompositePart>)partI.first(); cp != NULL;
	     cp = (Link<CompositePart>)partI.next())
	{
            if (debugMode) {
                printf("    In Reorg2, partId = %d:\n", cp->id);
            }
            count += cp->reorg2();
        }
    }
    return count;
}

//////////////////////////////////////////////////////////////////
//
// Composite Part Methods for Reorg2
//
//////////////////////////////////////////////////////////////////


o_4b CompositePart::reorg2()
{
    PartIdList 		toList;
    PartIdList 		fromList;
    AtomicPart* 	victim;
    Link<AtomicPart>	atomicPart;

    dirty(); 

    if (debugMode) {
        printf("CompositePart::Reorg2(id = %d)\n", id);
    }

    // first select and locate a victim atomic part

    o_4b	lowId = rootPart->id;
    o_4b	victimId = lowId + (o_4b) (random() % NumAtomicPerComp);

    LinkVstr<AtomicPart>	victims =
		PClassObject<AtomicPart>::Object().select(
			NULL, FALSE, PAttribute("AtomicPart:id") == victimId);

    victim = victims[0];
    victims.release();
    if ( victim == NULL ) {
	fprintf(stderr, "ERROR: Unable to find atomic part %d\n", victimId);
	exit(1);
    }

    // next record the affected to and from connections

    // first do the to connections
    o_4b	size = victim->to.size();
    o_word	i;

    for (i = 0; i < size; i++) {
	toList.insert(victim->to[i]->to->id);
    }

    // then do the from connections
    size = victim->from.size();
    for (i = 0; i < size; i++) {
	fromList.insert(victim->from[i]->from->id);
    }

    // now do the actual part deletion

    if (debugMode) {
        printf("\tReorg2: Deleting atomic part %d\n", victim->id);
    }

    // time to (approximately) recreate the deleted part
    atomicPart = new Persistent AtomicPart (victimId, this);

    if (victimId == lowId)
	rootPart = atomicPart;
    
    if (debugMode) {
        printf("\tReorg2: Recreated atomic part %d\n", atomicPart->id);
    }

    // finally, connect the part back up to the other parts again

    o_4b			partId;
    Link<Connection> 		connection;
    LinkVstr<AtomicPart>	atomicParts;

    while (!toList.empty()) {
	partId = toList.remove();
	atomicParts = PClassObject<AtomicPart>::Object().select(
			NULL, FALSE, PAttribute("AtomicPart:id") == partId);
	connection = new Persistent Connection(atomicPart, atomicParts[0]);
	connection->unpinobj();
	atomicParts.release();
    }

    while (!fromList.empty()) {
	partId = fromList.remove();
	atomicParts = PClassObject<AtomicPart>::Object().select(
			NULL, FALSE, PAttribute("AtomicPart:id") == partId);
	connection = new Persistent Connection(atomicParts[0], atomicPart);
	connection->unpinobj();
	atomicParts.release();
    }

    return 1;
}
